@@ -11,49 +11,48 @@ import "syscall"
 const maxSendfileSize int = 4 << 20
 
 // SendFile wraps the sendfile system call.
-func SendFile(dstFD *FD, src int, remain int64) (int64, error, bool) {
+func SendFile(dstFD *FD, src int, remain int64) (written int64, err error, handled bool) {
+	defer func() {
+		TestHookDidSendFile(dstFD, src, written, err, handled)
+	}()
 	if err := dstFD.writeLock(); err != nil {
 		return 0, err, false
 	}
 	defer dstFD.writeUnlock()
+
 	if err := dstFD.pd.prepareWrite(dstFD.isFile); err != nil {
 		return 0, err, false
 	}
 
 	dst := dstFD.Sysfd
-	var (
-		written int64
-		err     error
-		handled = true
-	)
 	for remain > 0 {
 		n := maxSendfileSize
 		if int64(n) > remain {
 			n = int(remain)
 		}
-		n, err1 := syscall.Sendfile(dst, src, nil, n)
+		n, err = syscall.Sendfile(dst, src, nil, n)
 		if n > 0 {
 			written += int64(n)
 			remain -= int64(n)
-		} else if n == 0 && err1 == nil {
-			break
-		}
-		if err1 == syscall.EINTR {
 			continue
-		}
-		if err1 == syscall.EAGAIN {
-			if err1 = dstFD.pd.waitWrite(dstFD.isFile); err1 == nil {
-				continue
-			}
-		}
-		if err1 != nil {
+		} else if err != syscall.EAGAIN && err != syscall.EINTR {
 			// This includes syscall.ENOSYS (no kernel
 			// support) and syscall.EINVAL (fd types which
-			// don't implement sendfile)
-			err = err1
-			handled = false
+			// don't implement sendfile), and other errors.
+			// We should end the loop when there is no error
+			// returned from sendfile(2) or it is not a retryable error.
+			break
+		}
+		if err == syscall.EINTR {
+			continue
+		}
+		if err = dstFD.pd.waitWrite(dstFD.isFile); err != nil {
 			break
 		}
 	}
-	return written, err, handled
+	if err == syscall.EAGAIN {
+		err = nil
+	}
+	handled = written != 0 || (err != syscall.ENOSYS && err != syscall.EINVAL)
+	return
 }
